import numpy
import pandas as pd
from pandas._testing import assert_frame_equal
from snowflake.sqlalchemy.snowdialect import ischema_names as snowflake_to_sql_alchemy_types


def snowflake_type_to_dtype(tname):
    return numpy.dtype(snowflake_type_to_sqltype(tname).python_type)


def snowflake_type_to_sqltype(tname):
    return snowflake_to_sql_alchemy_types.get(tname.strip())()


def process_cells(col_name_sqltype_pairs, cells):
    for col_name_sqltype_pair, cell in zip(col_name_sqltype_pairs, cells):
        cell = cell.strip()
        sql_type = col_name_sqltype_pair[1]
        if cell[0] == '"' and cell[-1] == '"':
            value = cell[1:-1]
        elif cell == "{null}":
            value = None
        else:
            value = cell
        if value is not None:
            value = value.lower() == "true" if sql_type.python_type is bool else sql_type.python_type(value)
        yield value


def table_to_df(table):
    heading = table.split("\n")[0]
    col_names_with_types = list(
        map(lambda header: header.strip(), heading.split("|")[1:-1]))
    col_name_sqltype_pairs = []
    for col_name_with_type in col_names_with_types:
        if len(col_name_with_type.split(':')) != 2:
            raise ValueError(
                f"You must specify name AND data type for columns like this 'my_field:string' at {col_name_with_type}")
        col_name, col_type = col_name_with_type.split(':')
        col_name_sqltype_pairs.append(
            (col_name.strip(), snowflake_type_to_sqltype(col_type.strip())))
    table_body = table.split("\n")[1:]
    table_body = list(filter(lambda x: "|" in x, table_body))
    rows = [list(process_cells(col_name_sqltype_pairs, row.split("|")[1:-1])) for row in table_body]

    df = pd.DataFrame(
        rows,
        columns=[col_name_sqltype_pair[0]
                 for col_name_sqltype_pair in col_name_sqltype_pairs],
    )

    return df, col_name_sqltype_pairs


def assert_frame_equal_with_sort(results, expected, key_columns, dtype_check=True):
    results_sorted = results.sort_values(
        by=key_columns).reset_index(drop=True).sort_index(axis=1)
    expected_sorted = expected.sort_values(
        by=key_columns).reset_index(drop=True).sort_index(axis=1)
    assert_frame_equal(results_sorted, expected_sorted,
                       check_index_type=False, check_dtype=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


gh_run_number = os.environ.get("BUILD_NUMBER", None)
build_number = None if gh_run_number is None or gh_run_number == "" else gh_run_number

version = '0.1.3'

setup(
    name='pytest-snowflake_bdd',
    version=f"{version}-{build_number}" if build_number else version,
    author='Tilak Patidar',
    author_email='tilakpatidar@gmail.com',
    maintainer='Tilak Patidar',
    maintainer_email='tilakpatidar@gmail.com',
    license='MIT',
    url='https://github.com/tilakpatidar/pytest-snowflake_bdd',
    description='Setup test data and run tests on snowflake in BDD style!',
    long_description=read('README.rst'),
    py_modules=['pytest_snowflake_bdd'],
    python_requires='>=3.6.7',
    install_requires=['pytest>=6.2.0', 'pytest-bdd>=3.2.1', 'snowflake-sqlalchemy>=1.3.2', 'SQLAlchemy>=1.4.27', \
                      'pandas>=0.25.3'],
    tests_require=[
      'tox',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Topic :: Database',
        'Topic :: Software Development :: Testing :: BDD',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: Pytest',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    packages=["pytest_snowflake_bdd"],
    entry_points={
        'pytest11': [
            'pytest-snowflake-bdd = pytest_snowflake_bdd.plugin',
        ],
    },
)



# -*- coding: utf-8 -*-
"""Pytest plugin entry point. Used for any fixtures needed."""

import pandas as pd
import pytest
from pytest_bdd import then, when, parsers, given
from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine, Column, MetaData, Table

from .utils import table_to_df, assert_frame_equal_with_sort


def pytest_addoption(parser):
    parser.addoption('--snowflake-user', required=True,
                     action='store', help='snowflake user for test environment')
    parser.addoption('--snowflake-password', required=True, action='store',
                     help='snowflake password for test environment')
    parser.addoption('--snowflake-account', required=True, action='store',
                     help='snowflake password for test environment')
    parser.addoption('--snowflake-role', required=False, action='store',
                     help='optional snowflake role for test environment',
                     default=None)
    parser.addoption('--snowflake-warehouse', required=False, action='store',
                     help='optional snowflake warehouse for test environment',
                     default=None)


@pytest.fixture
def snowflake_user(request):
    return request.config.getoption('--snowflake-user')


@pytest.fixture
def snowflake_password(request):
    return request.config.getoption('--snowflake-password')


@pytest.fixture
def snowflake_account(request):
    return request.config.getoption('--snowflake-account')


@pytest.fixture
def snowflake_role(request):
    return request.config.getoption('--snowflake-role')


@pytest.fixture
def snowflake_warehouse(request):
    return request.config.getoption('--snowflake-warehouse')


@pytest.fixture(scope="function")
def snowflake_sqlalchemy_conn(snowflake_user, snowflake_password, snowflake_account,
                              snowflake_role, snowflake_warehouse):
    yield from _snowflake_sqlalchemy_conn(snowflake_user, snowflake_password, snowflake_account, snowflake_role,
                                          snowflake_warehouse)


def _snowflake_sqlalchemy_conn(snowflake_user, snowflake_password, snowflake_account, snowflake_role,
                               snowflake_warehouse):
    optional_params = {}
    if snowflake_role:
        optional_params["role"] = snowflake_role
    if snowflake_warehouse:
        optional_params["warehouse"] = snowflake_warehouse

    engine = create_engine(URL(
        account=snowflake_account,
        user=snowflake_user,
        password=snowflake_password,
        **optional_params
    ))
    connection = engine.connect()
    yield engine
    connection.close()
    engine.dispose()


@when(parsers.re('a temporary table called "(?P<table_name>.+)" has\s+(?P<table>[\s\S]+)'))
def temp_table_create_fixture(snowflake_sqlalchemy_conn, table_name, table):
    create_table_with_data(snowflake_sqlalchemy_conn, table,
                           table_name, temporary=True)


@when(parsers.re('a table called "(?P<table_name>.+)" has\s+(?P<table>[\s\S]+)'))
def table_create_fixture(snowflake_sqlalchemy_conn, table_name, table):
    create_table_with_data(snowflake_sqlalchemy_conn, table,
                           table_name, temporary=False)


def create_table_with_data(snowflake_sqlalchemy_conn, table, table_name, temporary):
    df, col_name_sqltype_pairs = table_to_df(table)

    assert len(table_name.split(".")) == 3, "Table name should be fully qualified ex: db_name.schema_name.table_name"

    db_name, schema_name, tb_name = table_name.split(".")
    cols = [
        Column(col_name, col_type)
        for col_name, col_type in col_name_sqltype_pairs
    ]

    pd.read_sql(f"USE DATABASE \"{db_name}\"", snowflake_sqlalchemy_conn)

    metadata = MetaData(bind=snowflake_sqlalchemy_conn)
    temp_table = Table(tb_name,
                       metadata,
                       *cols,
                       schema=schema_name,
                       prefixes=['TEMPORARY'] if temporary else None,
                       )
    temp_table.create(bind=snowflake_sqlalchemy_conn)

    df.to_sql(con=snowflake_sqlalchemy_conn,
              schema=schema_name,
              name=tb_name,
              if_exists="append",
              method="multi",
              index=False)


@then(parsers.re('a sql script "(?P<script_path>.+)" runs and the result is\n(?P<table>[\s\S]+)'))
def assert_table_contains(snowflake_sqlalchemy_conn, script_path, table):
    sql = open(script_path, "r").read()
    print("Executing query")
    print(sql)

    actual_df = _fetch_results(snowflake_sqlalchemy_conn, sql)

    expected_df, _ = table_to_df(table)

    print("\n\n\nEXPECTED schema")
    print(expected_df.dtypes)
    print("\n\n\nACTUAL schema")
    print(actual_df.dtypes)

    assert_frame_equal_with_sort(actual_df, expected_df, key_columns=list(actual_df))


def _fetch_results(snowflake_sqlalchemy_conn, sql):
    res = snowflake_sqlalchemy_conn.execute(sql)
    columns = [line[0] for line in res.cursor.description]
    actual_df = pd.DataFrame(res, columns=columns)
    return actual_df


@given('a snowflake connection')
def t():
    pass
